// Copyright (c) 2000 by the University of Southern California
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.
//

// Header file for grid-based adaptive fidelity algorithm 

#ifndef ns_caf_h
#define ns_caf_h

#include <assert.h>
#include "agent.h"
#include "node.h"
#include "cntable.h"

#define MAXQUEUE  1             // 1 = only my grid, 5 = all of my neighbos
#define GN_UPDATE_INTERVAL 10 	// how often to update neighbor list
#define CAF_STARTUP_JITTER 1.0	// secs to jitter start of periodic activity
                                // also says CAF only stays in this state for
                                // 1.0 second
#define CAF_NONSTART_JITTER 3.0
#define MIN_DISCOVERY_TIME 1    // min interval to send discovery
#define MAX_DISCOVERY_TIME 15   // max interval to send discovery
#define MIN_SELECT_TIME 5	// start to tell your neighbor that
#define MAX_SELECT_TIME 6	// you are the leader b/w, my need
				// to be set by apps. 
#define MAX_DISCOVERY   10      // send selection after 10 time try
#define MIN_LIFETIME    60
#define CAF_LEADER_JITTER 3
#define MIN_TURNOFFTIME 1

#define NMAX 10      // The max # of cluster head neighbors of a node
                     //might have



#define CLUSTER_STARTTIME  2    // 2 sec to stay in the clutering time

#define CAF_DISCOVER  0    // discovery message type
#define CAF_GATEWAY   1    // gateway surpress type
#define CAF_LEADGATEWAY 2 // lead gateway broadcast

#define STATECHANGEBIAS 1
#define DISCOVER_SPACE  0.2
#define DISCOVER_MAXTIMES 5
#define DISCOVER_STARTTIME 1
#define DUTY_STARTTIME 2

class CAFAgent;


// they are used for timeout events
typedef enum {
  CAF_MSG_DISCOVER, CAF_MSG_SELECT, CAF_MSG_DUTY, CAF_MSG_SLEEP, CAF_MSG_SPDUTY, CAF_MSG_GATEWAY
} CafMsgType;

typedef enum {
  CAF_FREE, CAF_CLUSTER, CAF_LEADER, CAF_SLEEP
} CafNodeState;

typedef enum {
  CAF_PRIMARY_GATEWAY, CAF_NON_PRIMARY_GATEWAY, CAF_NON_GATEWAY
} CafNodeGatewayState;

/*
 * data structure for exchanging existence message and selection msg
 * Then followed by the varied-length neighbor list.
 */

struct DiscoveryMsgHead {
    u_int32_t nid;  // this node id
    int32_t cid;  //  this node's cluster id
    u_int32_t state; // this node's current state
    u_int32_t enlt;  // this node's estimated node live time
    u_int32_t stime;   // estimated leaving cluster time
    u_int32_t nlen;    // how many cluser head neighbors of this node
    u_int32_t nhead[NMAX]; // neighbor list,
};

/*
 * Gateway selection Msg: from gateway to the leader
 */
struct GatewayMsgHead {
    u_int32_t nid;  // this node id
    int32_t cid;  //  this node's cluster id
    u_int32_t gstate; // this node's current gstate
    u_int32_t nlen;    // how many cluster that this node can be gateway
    u_int32_t nhead[NMAX]; // neighbor list,
};

/*
 * From leader to the gateway
 */
struct LeadGatewayMsgHead {
    u_int32_t nid;  // this node id
    int32_t cid;  //  this node's cluster id
    u_int32_t nlen;    // how many cluster that this class has
    u_int32_t nhead[2*NMAX]; // neighbor node, cluster pair
};



// caf header

struct hdr_caf {
	// needs to be extended

  // int seqno_;
        int type_;

	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_caf* access(const Packet* p) {
		return (hdr_caf*) p->access(offset_);
	}
};


// CAFTimer is used for discovery phase

class CAFDiscoverTimer : public TimerHandler {
public: 
	CAFDiscoverTimer(CAFAgent *a) : TimerHandler(), a_(a) { }
protected:
	virtual void expire(Event *);
	CAFAgent *a_;
};

// CAFSelectTimer is for the slecting phase

class CAFSelectTimer : public TimerHandler {
public:
        CAFSelectTimer(CAFAgent *a) : TimerHandler(), a_(a) { }
protected:
        virtual void expire(Event *);
        CAFAgent *a_;
};

// CAFDutyTimer is for duty cycle. It is the place of adaptive fidelity
// plays

class CAFDutyTimer : public TimerHandler {
public:
        CAFDutyTimer(CAFAgent *a) : TimerHandler(), a_(a) { }
protected:
        inline void expire(Event *);
        CAFAgent *a_;
};

// CAF GRtimer is for gateway reselection

class CAFGRTimer : public TimerHandler {
public:
        CAFGRTimer(CAFAgent *a) : TimerHandler(), a_(a) { }
protected:
        inline void expire(Event *);
        CAFAgent *a_;
};

// CAFSPDutyTimer is for special duty cycle. 
// plays

class CAFSPDutyTimer : public TimerHandler {
public:
        CAFSPDutyTimer(CAFAgent *a) : TimerHandler(), a_(a) { }
protected:
        inline void expire(Event *);
        CAFAgent *a_;
};


class CAFAgent : public Agent {
public:
	CAFAgent(nsaddr_t id);
	virtual void recv(Packet *, Handler *);
	void timeout(CafMsgType);
	//void select_timeout(int);

	u_int32_t nodeid() {return nid_;}
	double myenlt();
	int caf_nosleep_;
protected:
	int command(int argc, const char*const*argv);
	

	void node_on();
	void node_off();
	void duty_timeout();
	void send_discovery();
	void send_gatewayinfo();
	void makeUpDiscoveryMsg(Packet *p);
	void makeUpGatewayMsg(Packet *p);
	void processDiscoveryMsg(Packet *p);
	void processGatewayMsg(Packet *p);
	void gateway_build_adjacent();

	void schedule_wakeup();
	void formCluster();
       	bool selectGateway();
	u_int32_t  enlt_;

	double beacon_; /* beacon period */
	void setCAFstate(CafNodeState);
	int randomflag_;

	double discover_starttime_offset_;
	double leader_starttime_;
	int discover_starttime_;
	int cluster_starttime_;

	double cluster_starttime_offset_;

	

	CAFDiscoverTimer timer_;
	CAFSelectTimer stimer_;
	CAFDutyTimer dtimer_;
	CAFSPDutyTimer sptimer_;
	CAFGRTimer grtimer_;

	double mywaketime_;
	
	u_int32_t  nid_;  // node id
	int32_t   cid_;  // cluster id
	u_int32_t   nlen_;  // how many cluster head neighbors
	u_int32_t nhead_[NMAX];  // limit it to NMAX cluster head neighbors.

	CAFnTable *table_;  // neighbor table
	

	Node *thisnode; // the node object where this agent resides
	CafNodeState	state_;

	CafNodeGatewayState gstate_;  // caf gateway state

	u_int32_t adjclusters_;   // how many adjacent clusters
	int32_t adjclusterlist_[NMAX]; // who are them

	double   init_eng_;
	double   threshold_eng_;


};

/* assisting getting broadcast msg */

class CAFPartner : public Connector {
public:
        CAFPartner();
        void recv(Packet *p, Handler *h);
	
protected:
	int command(int argc, const char*const*argv);
        ns_addr_t here_;
	int cafagent_;
        int mask_;
        int shift_;
};



#endif
